"""
Identifier:     csst_common/io.py
Name:           io.py
Description:    IO module
Author:         Bo Zhang
Created:        2023-12-13
Modified-History:
    2023-12-10, Bo Zhang, created
    2023-12-15, Bo Zhang, added verify_checksum and append_header, add module header
    2023-12-16, Bo Zhang, fixed a bug in finding the first key when reformatting
"""
import warnings
from copy import deepcopy
from typing import Optional

from astropy.io import fits


def verify_checksum(file_path) -> bool:
    """
    Verify a .fits file via checksum.

    Return True if checksum is good.

    Parameters
    ----------
    file_path : str
        File path.

    References
    ----------
    https://docs.astropy.org/en/stable/io/fits/usage/verification.html#verification-using-the-fits-checksum-keyword-convention
    """
    with warnings.catch_warnings(record=True) as warning_list:
        # file_path = fits.util.get_testdata_filepath('checksum_false.fits')
        with fits.open(file_path, checksum=True):
            pass
        print(warning_list)
    return len(warning_list) == 0


def append_header(
    h1: fits.Header, h2: fits.Header, duplicates: str = "delete"
) -> fits.Header:
    """
    Append h2 to h1.

    Append fits headers, taken into account duplicated keywords.

    Parameters
    ----------
    h1 : fits.Header
        Original fits header.
    h2 : fits.Header
        Extended fits header.
    duplicates : str
        The operation for processing duplicates.
        "delete" to delete duplicated keywords in h1 and keep them in h2.
        "update" to update duplicated keywords in h1 and remove them from h2.

    Returns
    -------
    fits.Header
        The combined header.

    References
    ----------
    https://docs.astropy.org/en/stable/io/fits/usage/headers.html#comment-history-and-blank-keywords
    """
    # copy data
    original_h = deepcopy(h1)
    extended_h = deepcopy(h2)
    original_keys = tuple(original_h.keys())
    extended_keys = tuple(extended_h.keys())
    ignored_keys = ("", "COMMENT", "HISTORY")
    assert duplicates in ("delete", "update")
    if duplicates == "update":
        for k in extended_keys:
            # print(card.keyword)
            if k not in ignored_keys and k in original_keys:
                print(f"Update existing key *{k} in original fits.Header*")
                original_h.set(k, extended_h[k], extended_h.comments[k])
                extended_h.remove(k)
    elif duplicates == "delete":
        for k in extended_keys:
            # print(card.keyword)
            if k not in ignored_keys and k in original_keys:
                print(f"Delete existing key *{k} in original fits.Header*")
                original_h.remove(k)
    original_h.extend(extended_h, bottom=True)
    return original_h


def reformat_header(
    h: fits.Header,
    strip: bool = True,
    comment: Optional[str] = None,
):
    """
    Reformat fits Header.

    Remove blanks, comments, and history, and add a new comment at the beginning.

    Parameters
    ----------
    h : fits.Header
        The original header.
    strip: bool = True
        If `True`, strip cards like SIMPLE, BITPIX, etc.
        so the rest of the header can be used to reconstruct another kind of header.
    comment: Optional[str] = None
        The new comment at the beginning.

    See Also
    --------
    astropy.io.fits.Header.strip()

    References
    ----------
    https://docs.astropy.org/en/stable/io/fits/api/headers.html#astropy.io.fits.Header.strip
    """
    h_copy = deepcopy(h)
    # strip
    if strip:
        h_copy.strip()
    # remove blanks, comments, and history
    h_copy.remove("", ignore_missing=True, remove_all=True)
    h_copy.remove("COMMENT", ignore_missing=True, remove_all=True)
    h_copy.remove("HISTORY", ignore_missing=True, remove_all=True)
    # add new comment
    assert len(h.cards) > 0
    first_key = list(h_copy.keys())[0]
    h_copy.add_comment("=" * 72, before=first_key)
    h_copy.add_comment(comment, before=first_key)
    h_copy.add_comment("=" * 72, before=first_key)
    return h_copy
